part of '../export.dart';

/// A Wave with a synchronously available and changeable value. Meant to reduce Signal boilerplate
///
/// - It doesn't implement Sink to suppress the close-sinks lint warning. (A signal does)
///
/// [Sea] is also a [Wave] so that it is backwards compatible with consumers
/// expecting a [Wave].
///
/// Add is only meant to be called inside of HookBlocs. That's why it's set
/// as protected.
class Sea<V> extends Wave<V>
    with WaveType<V>, WaveOtherMixin<V>, WaveOperators<V> {
  final Future<void> Function() _close;
  final Wave<V> wave;

  /// .add is supposed to be called in HookBloc subclasses never outside.
  /// You're supposed to add this line to the top of your HookBloc.
  /// You might also want to ignore type declarations because they get inferred.
  // ignore_for_file: invalid_use_of_protected_member
  // ignore_for_file: type_annotate_public_apis
  @protected
  final void Function(V) add;
  final V Function() __value;

  factory Sea(V s) {
    final signal = Signal(s);
    return Sea._create(
          () async => signal.close(),
      signal.wave,
      signal.add,
          () => signal.value,
    );
  }

  Sea._create(this._close, this.wave, this.add, this.__value) : super(wave);

  @override
  V get value => __value();

  Sea<V> operator &(
      Sea<V> Function(Sea<V> self, void Function() dispose) action) {
    return action(this, _close);
  }

  void change(V Function(V) action) => add(action(value));
}

/// Experimental High level HookBloc extensions for boilerplate free Blocs.
extension SeaHookBlocExtension on HookBloc {
  T d<T>(T t, void Function() dispose) {
    onDispose.add(() async => dispose());
    return t;
  }
}

extension SeaExtension<T> on T {
  /// Use this inside of [HookBloc]s as a replacement for [HookBloc.disposeSink].
  Sea<T> get $ {
    final sea = Sea(this);
    HookBloc.disposeEffect(sea, (Sea<T> a) async => a._close());
    return sea;
  }

  /// This is meant to be use when initialized inside of constructors.
  /// Use the & operator and [d] extension method on [HookBloc]s to dispose
  /// this [Sea].
  ///
  /// ```
  /// final a = 1.sea & d;
  /// ```
  ///
  Sea<T> get sea {
    return Sea(this);
  }
}